# -*- coding: utf-8 -*-
# Copyright (C) 2006-2007  Vodafone España, S.A.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Wrapper around common.SIMCardConnection

It basically provides error control and more high-level operations. N-tier
folks can see this as a Business Logic class.
"""

__version__ = "$Rev: 937 $"

from twisted.internet import reactor, defer

from vmc.common.protocol import SIMCardConnection
import vmc.common.exceptions as ex
from vmc.common.persistent import Contact
from vmc.common.smsng.sms import (pdu_to_message, unpack_ucs2_bytes,
                                  pack_ucs2_bytes, check_if_ucs2)

DELAY = 20

class SIMCardConnAdapter(SIMCardConnection):
    """Wrapper around common.SIMCardConnection"""
    
    def get_all_contacts(self):
        """Returns a list with all the contacts in the SIM"""
        def get_em(auxdefer=None):
            d2 = auxdefer and auxdefer or defer.Deferred()
            d = SIMCardConnection._get_all_contacts(self)
            def callback(contacts):
                clist = []
                for contact in contacts:
                    name = contact.group('name')
                    clist.append(Contact(name,
                                         contact.group('number'),
                                         int(contact.group('id'))))
                
                d2.callback(clist)
            
            def no_contacts_eb(failure):
                failure.trap(ex.CMEErrorNotFound, ex.ATError)
                d2.callback([])
            
            def sim_busy_eb(failure):
                failure.trap(ex.CMEErrorSIMBusy, ex.CMEErrorSIMNotStarted)
                # This basically is a loop that will be repeated until we get
                # a response. The reason is that some times the card will
                # reply with a SIM BUSY error or similar while trying to read
                # the contacts/messages, this is buggy and needs to be handled
                reactor.callLater(DELAY, get_em, d2)
            
            d.addCallback(callback)
            d.addErrback(no_contacts_eb)
            d.addErrback(sim_busy_eb)
        
            return d2
        
        return get_em()
    
    def get_all_sms(self):
        """
        Returns a list of ShortMessage objects with all the SMS in the SIM card
        """
        def get_em(auxdefer=None):
            d2 = auxdefer and auxdefer or defer.Deferred()
            
            d = SIMCardConnection.get_all_sms(self)
            def callback(messages):
                mlist = []
                for rawsms in messages:
                    # ShortMessage obj
                    sms = pdu_to_message(rawsms.group('pdu'))
                    sms.index = int(rawsms.group('id'))
                    sms.where = int(rawsms.group('storedat'))
                    mlist.append(sms)
                
                d2.callback(mlist)
            
            def error_sim_eb(failure):
                failure.trap(ex.CMEErrorNotFound, ex.ATError)
                d2.callback([])
            
            def sms_sim_busy_eb(failure):
                failure.trap(ex.CMEErrorSIMBusy,
                             ex.CMEErrorSIMNotStarted, ex.CMSError500)
                reactor.callLater(DELAY, get_em, d2)
            
            d.addCallback(callback)
            d.addErrback(error_sim_eb)
            d.addErrback(sms_sim_busy_eb)
            
            return d2
        
        return get_em()
    
    def get_available_character_set(self):
        """
        Returns a deferred that will be callbacked with the available
        character sets
        """
        d = SIMCardConnection.get_available_character_set(self)
        def process_callback(response):
            raw = response[0].group('langs')
            return raw.replace('"', '').split(',')
            
        d.addCallback(process_callback)
        return d
    
    def get_card_model(self):
        """Returns a deferred that will be callbacked with the card's model"""
        d = SIMCardConnection.get_card_model(self)
        d.addCallback(lambda response: response[0].group('model'))
        return d
    
    def get_card_version(self):
        """
        Returns a deferred that will be callbacked with the card's version
        """
        d = SIMCardConnection.get_card_version(self)
        d.addCallback(lambda response: response[0].group('version'))
        return d
    
    def get_character_set(self):
        d = SIMCardConnection.get_character_set(self)
        d.addCallback(lambda response: response[0].group('lang'))
        return d
    
    def get_imei_number(self):
        """
        Returns a deferred that will be callbacked with the card's IMEI number
        """
        d = SIMCardConnection.get_imei_number(self)
        d.addCallback(lambda response: response[0].group('imei'))
        return d
    
    def get_imsi_number(self):
        """
        Returns a deferred that will be callbacked with the card's IMSI number
        """
        d = SIMCardConnection.get_imsi_number(self)
        d.addCallback(lambda response: response[0].group('imsi'))
        return d
    
    def get_phonebook_size(self):
        """
        Returns a deferred that will be callbacked with the phonebook size
        of the SIM card
        """
        def get_it(auxdefer=None):
            d2 = auxdefer and auxdefer or defer.Deferred()
            
            d = SIMCardConnection.get_phonebook_size(self)
            def callback(resp):
                size = int(resp[0].group('size'))
                d2.callback(size)
            
            def errback(failure):
                failure.trap(ex.ATError, ex.CMEErrorSIMBusy, ex.CMSError500)
                reactor.callLater(DELAY, get_it, d2)
            
            d.addCallback(callback)
            d.addErrback(errback)
            return d2
        
        return get_it()
    
    def get_pin_status(self):
        """
        Returns a deferred that will be callbacked with 1 if PIN auth is active
        """
        def get_it(auxdefer=None):
            d2 = auxdefer and auxdefer or defer.Deferred()
            
            d = SIMCardConnection._get_pin_status(self)
            
            def callback(resp):
                active = int(resp[0].group('status'))
                d2.callback(active)
                
            def busy_errback(failure):
                failure.trap(ex.CMEErrorSIMBusy, ex.CMEErrorSIMNotStarted)
                reactor.callLater(DELAY, get_it, d2)
            
            def pinreq_errback(failure):
                failure.trap(ex.CMEErrorSIMPINRequired)
                d2.callback(1)
                
            d.addCallback(callback)
            d.addErrback(pinreq_errback)
            d.addErrback(busy_errback)
            
            return d2
        
        return get_it()
    
    
    def get_sms(self, index):
        """
        Returns a deferred that will be callbacked with ShortMessage object
        representing the SMS at C{index}
        """
        d = SIMCardConnection.get_sms(self, index)
        def callback(rawsms):
            rawsms = rawsms[0]
            sms = pdu_to_message(rawsms.group('pdu'))
            sms.index = index
            sms.where = int(rawsms.group('storedat'))
            
            return sms
        d.addCallback(callback)
        return d
    
    def get_smsc(self):
        """
        Returns a deferred that will be callbacked with the SMSC number
        """
        d = SIMCardConnection.get_smsc(self)
        def smsc_callback(response):
            smsc = response[0].group('smsc')
            if not smsc.startswith('+'):
                if check_if_ucs2(smsc):
                    # the smsc is in UCS2 format
                    smsc = unpack_ucs2_bytes(smsc).encode('utf8')
            return smsc
        
        d.addCallback(smsc_callback)
        return d
    
    def _get_imsi_prefix(self):
        d = self.get_imsi_number()
        d.addCallback(lambda response: int(response[:5]))
        return d
    
    def get_network_info(self, process=True):
        """
        Returns a deferred that will be callbacked with the network info
        
        The response will be a tuple as (OperatorName, ConnectionType) if
        it returns a (None, None) that means that some error occurred while
        obtaining the info. The class that requested the info should take
        care of insisting before this problem. This method will convert
        numeric network IDs to alphanumeric.
        """
        SIMCardConnection.set_network_info_format(self)
        d = SIMCardConnection.get_network_info(self)
        def callback(netinfo):
            if netinfo:
                status = int(netinfo[0].group('status'))
                conn_type = (status == 0) and 'GPRS' or '3G'
                netname = netinfo[0].group('netname')
                if netname in ['Limited Service',
                        pack_ucs2_bytes('Limited Service')]:
                    return None, None
                
                # netname can be in UCS2, as a string, or as a network id
                if check_if_ucs2(netname):
                    return unpack_ucs2_bytes(netname), conn_type
                else:
                    # now can be either a string or a network id (5 digits)
                    try:
                        netname = int(netname)
                    except ValueError:
                        # we got a string ID
                        return netname, conn_type

                    if not process:
                        return netname, conn_type
                    
                    # we got a numeric id, lets convert it
                    from vmc.utils.utilities import get_network_name_from_id
                    from vmc.utils.utilities import country_to_code
                    def callback(response):
                        if not response:
                            return None, None
                            
                        operator = response[0]
                        country = response[1]
                        try:
                            code = country_to_code[country]
                        except KeyError:
                            code = country
                        
                        op_name = operator + ' ' + code.upper()
                        return op_name, conn_type
                    
                    return get_network_name_from_id(
                                            netname).addCallback(callback)
            else:
                return None, None
        
        d.addCallback(callback)
        return d
    
    def get_network_names(self):
        """
        Returns a deferred that will be callbacked with a list of
        C{NetworkObject}s
        """
        d = SIMCardConnection.get_network_names(self)
        d.addCallback(lambda resp: [NetworkOperator(*match.groups())
                                        for match in resp])
        return d
    
    def register_with_network(self):
        """
        Returns a deferred that will be callbacked with the network registration
        result
        
        The algorithm of registering with the network is as follows:
         1. Get the five first digits of the IMSI
         2. Get the numeric ID of the network we are currently registered with
         3. If the ID is equals to the five first digits of IMSI, return OK
         4. If the ID is different, we will query the network about the
            available operators and if we find an operator that matches with
            the five digits of the IMSI, that means we have found our network
            and we will register with it.
        
        The algorithm nonetheless, is missing two major points:
         1. International roaming
         2. What to do if we can't find in the area our network operator ID
        
        Both must be solved through Vodafone.
        """
        imsi_def = self._get_imsi_prefix()
        response_deferred = defer.Deferred()
        def check_if_necessary(imsi_prefix):
            # is it really necessary to try to register with a network?
            netinfo_def = self.get_network_info(process=False)
            def check_if_cb(netinfo):
                netname, conn_type = netinfo
                if isinstance(netname, int):
                    if netname == imsi_prefix:
                        # netid != imsi[:5]
                        d = self.get_network_info()
                        d.chainDeferred(response_deferred)
                    else:
                        # netid != imsi[:5]
                        def register_callback(net_obj):
                            d = SIMCardConnection.register_with_network(self,
                                                                net_obj.netid)
                            def callback(resp):
                                d2 = self.get_network_info()
                                d2.chainDeferred(response_deferred)
                            
                            d.addCallback(callback)
                        
                        self._get_netobj_to_register(imsi_prefix)\
                                .addCallback(register_callback)
                else:
                    response_deferred.callback(netinfo)
            
            netinfo_def.addCallback(check_if_cb)
        
        imsi_def.addCallback(check_if_necessary)
        return response_deferred
    
    def _get_netobj_to_register(self, prefix):
        d = self.get_network_names()
        def callback(net_objs):
            for net_obj in net_objs:
                if net_obj.netid == prefix:
                    return net_obj
            
            return None
        d.addCallback(callback)
        return d
    
    def set_smsc(self, smsc):
        return SIMCardConnection.set_smsc(self, pack_ucs2_bytes(smsc))

class BasicNetworkOperator(object):
    def __init__(self, netid=None):
        super(BasicNetworkOperator, self).__init__()
        self.netid = netid
    
    def __repr__(self):
        args = (id(self), str(self.netid))
        return '<BasicNetworkOperator at 0x%x %s>' % args
    
    def __eq__(self, o):
        return self.netid == o.netid
    
    def __ne__(self, o):
        return not self.__eq__(o)


class NetworkOperator(BasicNetworkOperator):
    """
    I represent a network operator in the mobile network
    """
    def __init__(self, stat, long_name, short_name, netid, rat):
        super(NetworkOperator, self).__init__()
        self.stat = stat
        self.long_name = long_name
        self.short_name = short_name
        self.netid = netid
        self.rat = rat
    
    def __repr__(self):
        args = (self.long_name, self.netid)
        return '<NetworkOperator "%s" netid: %d>' % args
